
/*****************************************************************************
	Bounded linear
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#define MAXINT INT_MAX

#include "linearb.h"

extern long N;		// current transaction number
extern long LBBSIZE;	// batch size

/***************************************************************************** 
linearb_type *linearb_init(double rho, double epsilon, long M)
INPUTS:
	rho - hCount parameter
	epsilon - hCount parameter
	M - largest item in the stream

OUTPUT:
	Pointer to initialized linearb_type struct

Creates an instance of the linearb_type struct and initializes everything
*****************************************************************************/
linearb_type *linearb_init(double rho, double epsilon, long M)
{
    double e;			// min x/ln(x) and x > 1  --> evaluates to the base of the natural logarithm
    int R_RANGE;		// range for random numbers a_i and b_i
   
        int i;
    SLB_TYPE *aptr;

    linearb_type *result = calloc(1, sizeof(linearb_type));
    if (result == NULL) {
		printf("\ncalloc error");
		exit(100);
    }

    result->N = 0;			// total transactions
    result->M = M; 
    
	e = 2.718281828;
    //result->P = 1048583;	// big prime number
	//result->P = 1021;	
	result->P = 65539;
	//result->P = 16411;

    R_RANGE = result->M;
    
	result->m =  e / epsilon;		// length of hash table 
    result->h = (int) (log(-result->M / log(rho)));		// width of hash table
    
	// Allocate memory for the counters S
	aptr = (SLB_TYPE *) malloc(result->m * result->h * sizeof(SLB_TYPE));
    if (aptr == NULL) {
		printf("\nmalloc() error.");
		exit(100);
    }
    
	result->S = (SLB_TYPE **) malloc(result->m *sizeof(SLB_TYPE));
    if (result->S == NULL) {
		printf("\nmalloc() error.");
		exit(200);
    }
    for (i = 0; i < result->m; i++) {
		result->S[i] = aptr + i * result->h;
    }
    
	// Set all counters to 0
	memset((void *) aptr, 0, result->m * result->h * sizeof(SLB_TYPE) / sizeof(char));
    
	// generate h pairs of random numbers a_i and b_i
	result->a = (int *) malloc(result->h * sizeof(int));
    if (result->a == NULL) {
		printf("\nmalloc() error.");
		exit(300);
    }
    
	result->b = (int *) malloc(result->h * sizeof(int));
    if (result->b == NULL) {
		printf("\nmalloc() error.");
		exit(400);
    }
    for (i = 0; i < result->h; i++) {
		result->a[i] = random(0, R_RANGE);
		result->b[i] = random(0, R_RANGE);
    }

	return (result);
}

/*****************************************************************************
void linearb_update(win_type * hc, int k, int ttype)
INPUTS:
	hc - pointer to the hCount struct
	k - the item in the stream
	ttype - transaction type: 1 for INSERT, -1 for DELETE

OUTPUT:
	NIL

Updates the hCount table with new transaction. 
*****************************************************************************/
void linearb_update(linearb_type *hc, int val, int ttype)
{
        int pos,  j;
//	int carry, tmp;
//	int remainder;
	int w_id;

	N++;
	/* if (ttype == INSERT) {
		N++;
	} else {
		N--;
	} */

	w_id = N / LBBSIZE;

	/* Update the w_id level counters */
	for (j=0; j < hc->h; j++) {
		pos = ((((long long) hc->a[j] * val + hc->b[j]) % hc->P) % hc->m);
		hc->S[pos][j].c[w_id]++;
	}
}

/*****************************************************************************
long linearb_query(linearb_type *hc, long start, long end, long val)
INPUTS:
	hc - pointer to the hCount struct
	start - tqmin
	end - tqmax
	val - item to look for

OUTPUT:
	Count of "item" 

Counts the number of occurences of "item" in window (start, end)
*****************************************************************************/
long linearb_query(linearb_type *hc, long start, long end, long val)
{
        long min, pos,  j;
//	int powerValue;
        long finalCount;
//	int cnext = 0;
	long cnt;
	int wid_s, wid_e, wid;
	double frac_s, frac_e;

	wid_e = end / LBBSIZE;
	wid_s = start / LBBSIZE;

	frac_e = (double)(end - (wid_e*LBBSIZE))/(double)LBBSIZE;
	frac_s = (double)((wid_s+1)*LBBSIZE - start)/(double)LBBSIZE;

	if (wid_s == wid_e) {
		frac_s = (double)(end-start)/(double)LBBSIZE;
	}
		
	finalCount = 0;
	for (wid=wid_s; wid<=wid_e; wid++) {
		min = MAXINT;
		for (j=0; j < hc->h; j++) {
			pos = ((((long long) hc->a[j] * val + hc->b[j]) % hc->P) % hc->m);

			cnt = hc->S[pos][j].c[wid];
			if (cnt < min) {
				min = cnt;
			}
		}
		
		if (wid_s == wid_e) {
			min *= frac_s;
		} else {
			if (wid == wid_s) {
				min *= frac_s;
			}
			if (wid == wid_e) {
				min *= frac_e;
			}
		}
		
		finalCount += min;
	
	}

	return(finalCount);
}



/*****************************************************************************
int linearb_size(win_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	Space occupied by the hCount structure in bytes
*****************************************************************************/
int linearb_size(linearb_type * hc)
{
    int size;
    size = hc->m * hc->h * sizeof(SLB_TYPE);
    return (size);
}


/*****************************************************************************
void linearb_print(win_type *hc, int id)
INPUTS:
	hc - pointer to the hCount structure
	id - ID of Tilted Time Window

OUTPUT:
	NIL

Prints relevant debug information about the hCount structure
*****************************************************************************/
void linearb_print(linearb_type *hc, int id)
{
    int i, j;

    for (i = 0; i < hc->m; i++) {
		printf("\n");
		for (j = 0; j < hc->h; j++)
			printf("%d ", hc->S[i][j].c[id]);
    }

    for (i = 0; i < hc->h; i++) {
		printf("\n%d, a=%d, b=%d", i, hc->a[i], hc->b[i]);
    }

    printf("\nm = %d", hc->m);
    printf("\nh = %d", hc->h);
    printf("\nP = %d", hc->P);
    printf("\nwin_Print -- END");

}

/*****************************************************************************
void linearb_destroy(win_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	NIL

Frees up memory used by the hCount structure
*****************************************************************************/
void linearb_destroy(linearb_type * hc)
{
    free(hc->a);
    free(hc->b);
    free(hc->S);
}

